#!/bin/bash
set -e
km3pipe --help
pipeinspector --help
h5extract --help
h5extractf --help
h5info --help
h5tree --help
h5header --help
ztplot --help
k40calib --help
triggermap --help
nb2sphx --help
ligiermirror --help
qrunprocessor --help
qrunqaqc --help
daqsample --help
tres --help
